local helper = wesnoth.require "lua/helper.lua"
local T = helper.set_wml_tag_metatable {}

function wesnoth.get_mana(unit)
	-- A simple function to read the mana of a unit - this is not WML, so unit.variables.mana does not work
	local mana
	local max_mana
	for i = 1,#unit do
		if unit[i][1] == "variables" then
			if unit[i][2].mana and unit[i][2].max_mana then
				mana = tonumber(unit[i][2].mana)
				max_mana = tonumber(unit[i][2].max_mana)
			else
				helper.wml_error("Something is wrong with mana.")
			end
		end
	end
	return mana,max_mana
end

function wesnoth.display_mana(unit)
	-- When a unit's mana changes, its visual representation has to be updated, both on the mana bar and on the right panel, where it is represented as a trait, and disable or enable attacks that cost mana
	local t = {}
	-- Find and delete the overlay that marks the amount of mana the unit has
	for w in string.gmatch(unit.overlays, "[^%s,][^,]*") do
		if string.find(w, "mana_") then
		else
			table.insert(t, w)
		end
	end
	-- Calculate the percentage of mana the unit has
	local mana,max_mana = wesnoth.get_mana(unit)
	if mana > max_mana then
		mana = max_mana
	end
	local mana_percentage = mana / max_mana * 100
	mana_percentage = mana_percentage - mana_percentage % 5
	mana = mana - mana % 1
	max_mana = max_mana - max_mana % 1
	-- Add the corresponding picture to the overlays
	table.insert(t, "misc/mana/mana_" .. mana_percentage .. ".png")
	unit.overlays = table.concat(t, ',')
	-- Now we have to handle other things the mana affects - attacks and traits
	for i = 1,#unit do
		if unit[i][1] == "modifications" then
			-- Other traits don't belong here - we can delete them freely, they would occupy the traits panel uselessly, and we need it to represent more important stuff
			for j=1,#unit[i][2] do
				if unit[i][2][j] then
					if unit[i][2][j][1] == "trait" then
						table.remove(unit[i][2],j)
						unit[i][2][j] = nil
						j = j - 1
					end
				end
			end
			-- The number of attacks is not normally shown anywhere, so this trait might show it as well
			table.insert(unit[i][2] , { "trait" , { id="base_info", name="Mana: " .. mana .. "/" .. max_mana .. ", A: " .. unit.attacks_left .. "/" .. unit.max_attacks , } } )
		end
		if unit[i][1] == "attack" then
			-- Disable or enable attacks that require mana accordingly to the amount of mana the unit has
			local check
			for j=1,#unit[i][2] do
				if unit[i][2][j][1] == "specials" then
					for k=1,#unit[i][2][j][2] do
						if unit[i][2][j][2][k][1] == "chance_to_hit" then
							if unit[i][2][j][2][k][2].mana then
								if unit[i][2][j][2][k][2].mana > mana then
									check = "disable"
								else
									check = "enable"
								end		
							end
						end
					end
				end
			end
			if check == "disable" then
				unit[i][2].attack_weight = 0
			end
			if check == "enable" then
				unit[i][2].attack_weight = nil
			end
		end
	end
	return unit
end

function wesnoth.wml_actions.display_mana(cfg)
	-- WML function to use display_mana easily from WML
	for i,ru in ipairs(wesnoth.get_units(cfg)) do
		local u = ru.__cfg
		u = wesnoth.display_mana(u)
		wesnoth.put_unit(u)
	end
end

function wesnoth.calculate_max_experience(level)
	-- This formula might be changed, and is used on many locations, so it's better if its code is on a single location
	local max_experience =50 * level ^ 2 + 1.5 ^ level
	max_experience = max_experience - max_experience % 1
	return max_experience
end

function wesnoth.get_unit_properties(unit)
	-- This is the main function to gather information about the unit from the [a] tags that describe it.
	local race = "human"
	local intelligence = 10
	local strength = 10
	local dexterity = 10
	local endurance = 10
	local total_level = 0
	local melee_damage = 0
	local melee_damage_plus = 0
	local ranged_damage = 0
	local ranged_damage_plus = 0
	local melee_precision = 0
	local ranged_precision = 0
	local melee_attacks = 0
	local melee_attacks_plus = 0
	local ranged_attacks = 0
	local ranged_attacks_plus = 0
	local movement = 0
	local resist_fire = 0
	local resist_cold = 0
	local resist_arcane = 0
	local resist_blade = 0
	local resist_impact = 0
	local resist_pierce = 0
	local skills = {}
	local skills_modified = {}
	local specials_melee = {}
	local specials_ranged = {}
	local defense = 0
	local absorb = 0
	local to_all_skills = 0
	local melee_suck = 0
	local melee_suck_mana = 0
	local ranged_suck = 0
	local ranged_suck_mana = 0
	local penetrate_fire = 0
	local penetrate_cold = 0
	local penetrate_arcane = 0
	local penetrate_blade = 0
	local penetrate_impact = 0
	local penetrate_pierce = 0
	local magic_damage = 0
	local melee_type = "blade"
	local ranged_type = "pierce"
	local mana_regen = 0
	local hitpoints = 0
	local mana = 0
	local courage = false
	local effects = {}
	local alignment = "neutral"
	local class_levels = {}
	local str_req = 0
	local dex_req = 0
	local int_req = 0
	local melee_weapons_count = 0		-- We need to check if the player is dual-wielding or not, and add some penalties for it
	local skip_melee_weapons_count = false
	local weapon_type

	-- Items have to be calculated in a particularly complex way, so it needs a separate function, similar to describe_item, but there are differences in handling modifiers occasionally
	local function add_item_modifiers(item_code)
		if item_code.defense then
			defense = defense + item_code.defense
		end
		if item_code.courage then
			courage = true
		end
		if item_code.alignment then
			alignment = item_code.alignment
		end
		if item_code.movement then
			movement = movement + item_code.movement
		end
		if item_code.absorb then
			absorb = absorb + item_code.absorb
		end
		if item_code.resist_fire then
			resist_fire = resist_fire + item_code.resist_fire
		end
		if item_code.resist_cold then
			resist_cold = resist_cold + item_code.resist_cold
		end
		if item_code.resist_arcane then
			resist_arcane = resist_arcane + item_code.resist_arcane
		end
		if item_code.resist_pierce then
			resist_pierce = resist_pierce + item_code.resist_pierce
		end
		if item_code.resist_blade then
			resist_blade = resist_blade + item_code.resist_blade
		end
		if item_code.resist_impact then
			resist_impact = resist_impact + item_code.resist_impact
		end
		if item_code.increase_strength then
			strength = strength + item_code.increase_strength
		end
		if item_code.increase_dexterity then
			dexterity = dexterity + item_code.increase_dexterity
		end
		if item_code.increase_endurance then
			endurance = endurance + item_code.increase_endurance
		end
		if item_code.increase_intelligence then
			intelligence = intelligence + item_code.increase_intelligence
		end
		if item_code.mana_regen then
			mana_regen = mana_regen + item_code.mana_regen
		end
		if item_code.magic then
			magic_damage = magic_damage + item_code.magic
		end
		if item_code.mana then
			mana = mana + item_code.mana
		end
		if item_code.hitpoints then
			hitpoints = hitpoints + item_code.hitpoints
		end
		if item_code.to_all_skills then
			to_all_skills = to_all_skills + item_code.to_all_skills
		end
		if item_code.weapon_type then
			weapon_type = item_code.weapon_type
		end
		if item_code.penetrate_fire then
			penetrate_fire = penetrate_fire + item_code.penetrate_fire
		end
		if item_code.penetrate_cold then
			penetrate_cold = penetrate_cold + item_code.penetrate_cold
		end
		if item_code.penetrate_arcane then
			penetrate_arcane = penetrate_arcane + item_code.penetrate_arcane
		end
		if item_code.penetrate_blade then
			penetrate_blade = penetrate_blade + item_code.penetrate_blade
		end
		if item_code.penetrate_pierce then
			penetrate_pierce = penetrate_pierce + item_code.penetrate_pierce
		end
		if item_code.penetrate_impact then
			penetrate_impact = penetrate_impact + item_code.penetrate_impact
		end
		if item_code.required_strength then
			if item_code.required_strength > str_req then
				str_req = item_code.required_strength
			end
		end
		if item_code.required_dexterity then
			if item_code.required_dexterity > dex_req then
				dex_req = item_code.required_dexterity
			end
		end
		if item_code.required_intelligence then
			if item_code.required_intelligence > int_req then
				int_req = item_code.required_intelligence
			end
		end
		-- Melee weapons should not affect ranged damage and vice versa (except attribute bonuses)
		if item_code.sort ~= "melee" and item_code.sort ~= "ranged" and item_code.sort ~= "wand" then
			if item_code.damage then
				melee_damage = melee_damage + item_code.damage
				ranged_damage = ranged_damage + item_code.damage
			end
			if item_code.damage_plus then
				melee_damage_plus = melee_damage_plus + item_code.damage_plus
				ranged_damage_plus = ranged_damage_plus + item_code.damage_plus
			end
			if item_code.attacks then
				melee_attacks = melee_attacks + item_code.attacks
				ranged_attacks = ranged_attacks + item_code.attacks
			end
			if item_code.attacks_plus then
				melee_attacks_plus = melee_attacks_plus + item_code.attacks_plus
				ranged_attacks_plus = ranged_attacks_plus + item_code.attacks_plus
			end
			if item_code.precision then
				melee_precision = melee_precision + item_code.precision
				ranged_precision = ranged_precision + item_code.precision
			end
			if item_code.suck then
				melee_suck = melee_suck + item_code.suck
				ranged_suck = ranged_suck + item_code.suck
			end
			if item_code.suck_mana then
				melee_suck_mana = melee_suck_mana + item_code.suck_mana
				ranged_suck_mana = ranged_suck_mana + item_code.suck_mana
			end
			if item_code.damage_type then
				melee_type = item_code.damage_type
				ranged_type = item_code.damage_type
			end
			for aim_i=1,#item_code do
				if item_code[aim_i][1] == "specials" then
					table.insert(specials_melee,item_code[aim_i][2])
					table.insert(specials_ranged,item_code[aim_i][2])
				end
			end
		end
		if item_code.sort == "melee" or item_code.sort == "wand" then
			if not skip_melee_weapons_count then
				melee_weapons_count = melee_weapons_count + 1
			end
			if item_code.damage then
				melee_damage = melee_damage + item_code.damage
			end
			if item_code.damage_plus then
				melee_damage_plus = melee_damage_plus + item_code.damage_plus
			end
			if item_code.attacks then
				melee_attacks = melee_attacks + item_code.attacks
			end
			if item_code.attacks_plus then
				melee_attacks_plus = melee_attacks_plus + item_code.attacks_plus
			end
			if item_code.precision then
				melee_precision = melee_precision + item_code.precision
			end
			if item_code.suck then
				melee_suck = melee_suck + item_code.suck
			end
			if item_code.suck_mana then
				melee_suck_mana = melee_suck_mana + item_code.suck_mana
			end
			if item_code.damage_type then
				melee_type = item_code.damage_type
			end
			for aim_i=1,#item_code do
				if item_code[aim_i][1] == "specials" then
					table.insert(specials_melee,item_code[aim_i][2])
				end
			end
		end
		if item_code.sort == "ranged" then
			if item_code.damage then
				ranged_damage = ranged_damage + item_code.damage
			end
			if item_code.damage_plus then
				ranged_damage_plus = ranged_damage_plus + item_code.damage_plus
			end
			if item_code.attacks then
				ranged_attacks = ranged_attacks + item_code.attacks
			end
			if item_code.attacks_plus then
				ranged_attacks_plus = ranged_attacks_plus + item_code.attacks_plus
			end
			if item_code.precision then
				ranged_precision = ranged_precision + item_code.precision
			end
			if item_code.suck then
				ranged_suck = ranged_suck + item_code.suck
			end
			if item_code.suck_mana then
				ranged_suck_mana = ranged_suck_mana + item_code.suck_mana
			end
			if item_code.damage_type then
				ranged_type = item_code.damage_type
			end
			for aim_i=1,#item_code do
				if item_code[aim_i][1] == "specials" then
					table.insert(specials_ranged,item_code[aim_i][2])
				end
			end
		end
		-- Subtags have to be handled differently
		for aim_i=1,#item_code do
			if item_code[aim_i][1] == "effect" then
				table.insert(effects,item_code[aim_i])
			end
			if item_code[aim_i][1] == "skills" then
				for j,k in pairs(item_code[aim_i][2]) do
					skills_modified[tonumber(j)] = k
				end
			end
		end

	end
	-- Easier access to the container that contains the [a] tags
	for i=1,#unit do
		if unit[i][1] == "variables" then
			properties = unit[i][2]
		end
	end
	-- Main analysis of [a] tags
	for i=1,#properties do
		if properties[i][1] == "a" then
			-- The first numbers indicate the race - there are no possible additional values, just one of them, so 30 possibilities is way more than enough
			if properties[i][2].id <= 30 then
				race = races[properties[i][2].id][2].name
				for j=1,#races[properties[i][2].id][2] do
					if races[properties[i][2].id][2][j][1] == "properties" then
						local item_code_picked = races[properties[i][2].id][2][j][2]
						add_item_modifiers(item_code_picked)
					end
				end
			end
			-- Attributes are only four and can't be normally increased much over 60 - 100 possible values should be enough
			if properties[i][2].id < 200 and properties[i][2].id >= 100 then
				strength = properties[i][2].id - 100
			end
			if properties[i][2].id < 300 and properties[i][2].id >= 200 then
				dexterity = properties[i][2].id - 200
			end
			if properties[i][2].id < 400 and properties[i][2].id >= 300 then
				endurance = properties[i][2].id - 300
			end
			if properties[i][2].id < 500 and properties[i][2].id >= 400 then
				intelligence = properties[i][2].id - 400
			end
			-- Numbers 1000 to 10000 are taken for crafting gems, that are not needed here, only in import/export things, there might be 20 of them and maximum quantity of each 50
			-- Racial skills are separated from classical skills, they don't need big numbers because they're not different for classes and there aren't meant to be too numerous, not all races even have one
			if properties[i][2].id <11000  and properties[i][2].id >= 10000 then
				local skill_level = (properties[i][2].id - 10000) % 10
				local skill_number = (properties[i][2].id - 10000 - skill_level) / 10
				skills[skill_number] = skill_level
			end
			-- Class levels - class skills are not enough to learn them, beause they might select racial skills instead of class skills when advancing, space for 30 classes and maximum level 99, that is
			-- more than enough
			if properties[i][2].id <16000  and properties[i][2].id >= 11000 then
				total_level = total_level + properties[i][2].id % 100
				class_levels[(properties[i][2].id - 11000 - properties[i][2].id % 100) / 100 ] = properties[i][2].id % 100
			end
			-- Class skills - there is space for 30 classes, each class has space for 100 skills, that should be enough even for all possible additional expansions
			if properties[i][2].id <60000  and properties[i][2].id >= 20000 then
				local skill_level = (properties[i][2].id - 20000) % 10
				local skill_number = (properties[i][2].id - skill_level - 20000) / 10
				skills[skill_number] = skill_level
			end
			-- Unique items - because they have no additional values, there is space for 10000 of them, that we are not very likely to make up
			if properties[i][2].id <70000  and properties[i][2].id >= 60000 then
				local item_code_picked = uniques[properties[i][2].id-60000][2]
				add_item_modifiers(item_code_picked)
				-- Will be used also for set items, that should be entirely pre-written as well
			end
			if properties[i][2].id <200000  and properties[i][2].id >= 100000 then
				-- Will be used for crafted items, the first three digits will indicate the modifiers added and the other two will represent the base item
			end
			-- Randomly generated items - the first two digits indicate the prefix, the other two represent the base item and the last couple represent the suffix, zeroes mean that the modifier is absent
			if properties[i][2].id <1000000  and properties[i][2].id >= 200000 then
				local base_item = base_items[((properties[i][2].id - 200000) % 10000 - (properties[i][2].id - 200000) % 100)/100][2] or helper.wml_error("Base item of" .. number .. "is not defined")
				add_item_modifiers(base_item)
				if properties[i][2].id > 210100 then
					local prefix = prefixes[(properties[i][2].id - 200000 - (properties[i][2].id - 200000) % 10000)/10000][2] or helper.wml_error("Prefix of" .. number .. "is not defined")
					-- It is essential to know on what kind of item the modifiers are, or the modifiers of melee weapon's prefix might affect a ranged weapon, and that would suck
					prefix.sort=base_item.sort
					skip_melee_weapons_count = true		-- Modifiers on melee weapons would count as additional weapons and create a suspicion for dual-wielding
					add_item_modifiers(prefix)
					skip_melee_weapons_count = false
					-- The change in the properties of the prefix has to be undone, or the thought-to-be unchangable data files read will contain misinformation, confusing wesnoth.describe_item
					prefix.sort = nil	
				end
				if properties[i][2].id % 100 > 0 then
					-- Same for suffix
					local suffix = suffixes[ properties[i][2].id % 100][2] or helper.wml_error("Suffix of" .. number .. "is not defined")
					suffix.sort=base_item.sort
					skip_melee_weapons_count = true
					add_item_modifiers(suffix)
					skip_melee_weapons_count = false
					suffix.sort = nil
				end
			end
			-- Numbers 1060000 to 1070000 are taken for unique items in inventory, that are not needed here, only in import/export things and inventory manipulation
			-- Numbers 1100000 to 1200000 are taken for crafted itemsin inventory, that are not needed here, only in import/export things and inventory manipulation
			-- Numbers 1200000 to 2000000 are taken for randomly generated itemsin inventory, that are not needed here, only in import/export things and inventory manipulation
			-- Numbers above 2000000 indicate the experience of units, and that is needed only in import/export things (and maybe some anti-cheating thing later)
		end
	end
	-- Dualwielding penalty - take a dagger with +150% attacks and a mace with +200% damage and watch how it stacks multiplicatively
	if melee_weapons_count == 2 then
		melee_precision = melee_precision - 10
		melee_damage = melee_damage / 1.5
		melee_damage_plus = melee_damage_plus / 1.5
	end
	-- Base skills read will be modified by skill bonuses from item
	for j,k in pairs(skills) do
		j = tonumber(j)
		if skills_modified[j] then
			skills_modified[j] = skills_modified[j] + k
		else
			skills_modified[j] = k
		end
	end
	if to_all_skills > 0 then	-- This precious modifier will modify only the skills the unit already has
		j = tonumber(j)
		for j,k in pairs(skills_modified) do
			skills_modified[j] = k + to_all_skills
		end
	end
	return race, intelligence, strength, dexterity, endurance, total_level, melee_damage, melee_damage_plus, ranged_damage, ranged_damage_plus, melee_precision, ranged_precision, melee_attacks, melee_attacks_plus, ranged_attacks, ranged_attacks_plus, movement, resist_fire, resist_cold, resist_arcane, resist_blade, resist_impact, resist_pierce, skills, specials_melee, specials_ranged, defense, absorb, skills_modified, melee_suck, melee_suck_mana, ranged_suck, ranged_suck_mana, penetrate_fire, penetrate_cold, penetrate_arcane, penetrate_blade, penetrate_impact, penetrate_pierce, magic_damage, melee_type, ranged_type, mana_regen, hitpoints, courage, effects, alignment, class_levels, mana, weapon_type, str_req, int_req, dex_req
end

function wesnoth.message_box(message)
	-- A small window that shows some announcement
	local message_box_dialog = { T.tooltip { id = "tooltip_large" }, T.helptip { id = "tooltip_large" },  maximum_height = 600, maximum_width = 800,
		T.grid { T.row { T.column { T.label { id="title" , definition="title" , label="Announcement"} }} ,
			T.row {	T.column { T.scroll_label { label=message} } },
			T.row { T.column { T.button { id = "cancel", label = "Okay...", tooltip="Back to game." } }}
		}}
	local function message_box_preshow()
	end
	local function message_box_postshow()
	end
	-- The decision has no effect on the game, but we don't want other players to see this message
	wesnoth.synchronize_choice(
		function()
			wesnoth.show_dialog(message_box_dialog,message_box_preshow,message_box_postshow)
		end
	)
end		

function wesnoth.wml_actions.damage_unit(cfg)
	-- Clone of harm_unit (even with comments, why not?), just gives experience if the target dies also if the experience=no, and shares the experience between all players
	local filter = helper.get_child(cfg, "filter") or helper.wml_error("[damage_unit] missing required [filter] tag")
	-- we need to use shallow_literal field, to avoid raising an error if $this_unit (not yet assigned) is used
	if not cfg.__shallow_literal.amount then helper.wml_error("[damage_unit] has missing required amount= attribute... or the WML code just failed to calculate it, maybe the formula is messed up") end
	local variable = cfg.variable -- kept out of the way to avoid problems
	local _ = wesnoth.textdomain "wesnoth"
	-- #textdomain wesnoth
	local harmer


	for index, unit_to_harm in ipairs(wesnoth.get_units(filter)) do
		if unit_to_harm.valid then
			wesnoth.set_variable ( "this_unit" ) -- clearing this_unit
			wesnoth.set_variable("this_unit", unit_to_harm.__cfg) -- cfg field needed
			local amount = tonumber(cfg.amount) or helper.wml_error("[damage_unit] has missing required amount= attribute")
			local animate = cfg.animate -- attacker and defender are special values
			local delay = cfg.delay or 500
			local kill = cfg.kill
			local fire_event = cfg.fire_event
			local primary_attack = helper.get_child(cfg, "primary_attack")
			local secondary_attack = helper.get_child(cfg, "secondary_attack")
			local harmer_filter = helper.get_child(cfg, "filter_second")
			local experience = cfg.experience
			if harmer_filter then harmer = wesnoth.get_units(harmer_filter)[1] end
			if animate then
				if animate ~= "defender" and harmer and harmer.valid then
					wesnoth.scroll_to_tile(harmer.x, harmer.y, true)
					wml_actions.animate_unit( { flag = "attack", hits = true, { "filter", { id = harmer.id } },
						{ "primary_attack", primary_attack },
						{ "secondary_attack", secondary_attack }, with_bars = true,
						{ "facing", { x = unit_to_harm.x, y = unit_to_harm.y } } } )
				end
				wesnoth.scroll_to_tile(unit_to_harm.x, unit_to_harm.y, true)
			end

			-- the two functions below are taken straight from the C++ engine, util.cpp and actions.cpp, with a few unuseful parts removed
			-- may be moved in helper.lua in 1.11
			local function round_damage( base_damage, bonus, divisor )
				local rounding
				if base_damage == 0 then return 0
				else
					if bonus < divisor or divisor == 1 then
						rounding = divisor / 2 - 0
					else
						rounding = divisor / 2 - 1
					end
					return math.max( 1, math.floor( ( base_damage * bonus + rounding ) / divisor ) )
				end
			end

			local function calculate_damage( base_damage, alignment, tod_bonus, resistance )
				local damage_multiplier = 100
				if alignment == "lawful" then
					damage_multiplier = damage_multiplier + tod_bonus
				elseif alignment == "chaotic" then
					damage_multiplier = damage_multiplier - tod_bonus
				elseif alignment == "liminal" then
					damage_multiplier = damage_multiplier - math.abs( tod_bonus )
				else -- neutral, do nothing
				end
				damage_multiplier = damage_multiplier * resistance -- at this point, a resistance_modifier can be added, as asked by fendrin
				local damage = round_damage( base_damage, damage_multiplier, 10000 ) -- if harmer.status.slowed, this may be 20000 ?
				return damage
			end
			local damage = calculate_damage( amount,
							 ( cfg.alignment or "neutral" ),
							 wesnoth.get_time_of_day( { unit_to_harm.x, unit_to_harm.y, true } ).lawful_bonus,
							 wesnoth.unit_resistance( unit_to_harm, cfg.damage_type or "dummy" )
						       )

			if unit_to_harm.hitpoints <= damage then
				if kill == false then damage = unit_to_harm.hitpoints - 1
				else damage = unit_to_harm.hitpoints
				end
			end

			unit_to_harm.hitpoints = unit_to_harm.hitpoints - damage
			local text = string.format("%d%s", damage, "\n")
			local add_tab = false
			local gender = unit_to_harm.__cfg.gender

			local function set_status(name, male_string, female_string, sound)
				if not cfg[name] or unit_to_harm.status[name] then return end
				if gender == "female" then
					text = string.format("%s%s%s", text, tostring(female_string), "\n")
				else
					text = string.format("%s%s%s", text, tostring(male_string), "\n")
				end

				unit_to_harm.status[name] = true
				add_tab = true

				if animate and sound then -- for unhealable, that has no sound
					wesnoth.play_sound (sound)
				end
			end

			if not unit_to_harm.status.not_living then
				set_status("poisoned", _"poisoned", _"female^poisoned", "poison.ogg")
			end
			set_status("slowed", _"slowed", _"female^slowed", "slowed.wav")
			set_status("petrified", _"petrified", _"female^petrified", "petrified.ogg")
			set_status("unhealable", _"unhealable", _"female^unhealable")

			if add_tab then
				text = string.format("%s%s", "\t", text)
			end

			if animate and animate ~= "attacker" then
				if harmer and harmer.valid then
					wml_actions.animate_unit( { flag = "defend", hits = true, { "filter", { id = unit_to_harm.id } },
						{ "primary_attack", primary_attack },
						{ "secondary_attack", secondary_attack }, with_bars = true },
						{ "facing", { x = harmer.x, y = harmer.y } } )
				else
					wml_actions.animate_unit( { flag = "defend", hits = true, { "filter", { id = unit_to_harm.id } },
						{ "primary_attack", primary_attack },
						{ "secondary_attack", secondary_attack }, with_bars = true } )
				end
			end

			wesnoth.float_label( unit_to_harm.x, unit_to_harm.y, string.format( "<span foreground='red'>%s</span>", text ) )

			local function calc_xp( level ) -- to calculate the experience in case of kill
				if level == 0 then return 4
				else return level * 8 end
			end
			local players = wesnoth.get_units { canrecruit=true }
			if harmer and harmer.valid and wesnoth.is_enemy( unit_to_harm.side, harmer.side ) then -- no XP earned for harming friendly units
				-- Fire the experience-sharing function instead of adding it normally - damage_unit specific
				if kill ~= false and unit_to_harm.hitpoints <= 0 then
					wesnoth.wml_actions.share_experience{ amount=math.ceil(calc_xp( unit_to_harm.__cfg.level ))}
				else
					if experience ~= false then
						unit_to_harm.experience = unit_to_harm.experience + harmer.__cfg.level
						harmer.experience = harmer.experience + unit_to_harm.__cfg.level
					end
				end
			end
			if (not harmer or not harmer.valid) and experience ~= false then	-- Give experience also if the harmer is unknown
				if kill ~= false and unit_to_harm.hitpoints <= 0 then
					wesnoth.wml_actions.share_experience{ amount=math.ceil(calc_xp( unit_to_harm.__cfg.level ))}
				end
			end
			if kill ~= false and unit_to_harm.hitpoints <= 0 then
				local function bool( value ) -- support function for kill tag below
					if value then return true
					else return false end
				end
				wesnoth.wml_actions.kill({ id = unit_to_harm.id, animate = bool( animate ), fire_event = fire_event })
			end

			if animate then
				wesnoth.delay(delay)
			end

			if variable then
				wesnoth.set_variable(string.format("%s[%d]", variable, index - 1), { harm_amount = damage })
			end

			-- both may no longer be alive at this point, so double check
			if  harmer and unit_to_harm.valid and unit_to_harm.experience >= unit_to_harm.max_experience then
				wml_actions.store_unit { { "filter", { id = unit_to_harm.id } }, variable = "Lua_store_unit", kill = true }
				wml_actions.unstore_unit { variable = "Lua_store_unit", find_vacant = false, advance = true }
				wesnoth.set_variable ( "Lua_store_unit", nil )
			end
		end

		wesnoth.wml_actions.redraw {}
	end
end

function wesnoth.wml_actions.share_experience(cfg)
	-- In some cases, the experience will have to be shared between all units, especially with AoE kills or summons
	local raw_players = wesnoth.get_units { canrecruit=true , side="1,2,3,4,5"}
	local players = raw_players[1].__cfg
	local amount = cfg.amount or helper.wml_error("[share_experience] is lacking the mandatory amount= variable")
	for j = 1,#raw_players do
		local player = raw_players[j].__cfg
		player.experience = player.experience + math.ceil(amount / #raw_players)
		wesnoth.put_unit(player)
	end
end

function wesnoth.wml_actions.reload_unit_modifications(cfg)
	-- This is a simple function helping WML - if we remove an [object] or something, we usually need to update its properies, but using [transform_unit] and restoring all other properties every time
	-- is tedious. It is simpler to make a WML tag for it.
	-- CAREFUL: Works only for normal units, it's way too simple for heroes, but in that case it might be simply done using wesnoth.update_unit
	for i,ru in ipairs(wesnoth.get_units(cfg)) do
		local u = ru.__cfg
		local health_percentage = u.hitpoints / u.max_hitpoints
		local experience_percentage = u.experience / u.max_experience
		local moves_percentage = u.moves / u.max_moves
		local status = helper.get_child( u, "status" )
		local variables = helper.get_child( u, "variables" )
		local modifications = helper.get_child( u, "modifications" )
		local raw_unit = wesnoth.create_unit { type=u.type, id=u.id, gender=u.gender, facing=u.facing, x=u.x, y=u.y, canrecruit=u.canrecruit, side=u.side, attacks_left=u.attacks_left, name=u.name, { "variables" , variables }, {"modifications" , modifications }, { "status" , status} }
		u = raw_unit.__cfg
		u.moves = u.max_moves * moves_percentage
		u.hitpoints = u.max_hitpoints * health_percentage
		u.experience = u.max_experience * experience_percentage
		wesnoth.put_unit(u)
	end
end

function wesnoth.wml_actions.calculate_mana()
	-- This function handles the mana costs of enchantments and summons the unit might have. It is invoked by WML at turn refresh. If there is not enough mana, the enchantments fail and summons vanish.
	-- It is simply handled by a variable named mana_to_pay, that is the amount of mana that the unit will lose each turn after that the mana is restored. Cooldowns are calculated here as well.
	local side_number = wesnoth.get_variable("side_number")
	local raw_units = wesnoth.get_units { side=side_number, {"filter_wml" , {  { "variables" , {  enchanted = true }}  }} }
	local raw_summons = wesnoth.get_units { side=side_number, {"filter_wml" , {  { "variables" , {  is_summon = true }}  }} }
	-- Summons should not collapse first, they can be quite good meatshields, and their destruction might cause more damage than the collapse of enchantments - so the mana for them is checked first, when it should
	-- be still available
	for i = 1,#raw_summons do
		local unit = raw_summons[i].__cfg
		local enchantments = false
		for j = 1,#unit do
			if unit[j][1] == "variables" then
				for k = 1,#unit[j][2] do
					-- Find the information about the summoning spell
					if unit[j][2][k][1] == "summon_data" then
						local raw_summoner = wesnoth.get_units { side=side_number, canrecruit= true }
						local summoner = raw_summoner[1].__cfg
						local mana_needed = unit[j][2][k][2].cost
						-- Make the summoner's mana_to_pay increase
						for l = 1,#summoner do
							if summoner[l][1] == "variables" then
								local mana_needed = unit[j][2][k][2].cost + summoner[l][2].mana_to_pay - summoner[l][2].mana_regen
								if summoner[l][2].mana >= mana_needed then
									summoner[l][2].mana_to_pay = summoner[l][2].mana_to_pay + unit[j][2][k][2].cost
								else
									-- Kill the summon if there is not enough mana
									wesnoth.wml_actions.kill{ id = unit.id, fire_event = false}
								end

							end
						end
						wesnoth.put_unit(summoner)
					end
				end
			end
		end
	end
	-- And take the mana for keeping enchantments
	for i = 1,#raw_units do
		local unit = raw_units[i].__cfg
		local enchantments = false
		for j = 1,#unit do
			if unit[j][1] == "variables" then
				for k = 1,#unit[j][2] do
					if unit[j][2][k] then
						if unit[j][2][k][1] == "enchantment" then
							-- Each enchantment should collapse if the enchanter does not have enough mana or cannot be found
							local raw_enchanter = wesnoth.get_units { side=unit[j][2][k][2].caster, canrecruit=true }
							local enchanter = raw_enchanter[1].__cfg
							if enchanter then
								local enchanter = raw_enchanter[1].__cfg
								local mana_needed = unit[j][2][k][2].cost
								for l = 1,#enchanter do
									if enchanter[l][1] == "variables" then
										local mana_needed = unit[j][2][k][2].cost + enchanter[l][2].mana_to_pay - enchanter[l][2].mana_regen
										if enchanter[l][2].mana >= mana_needed then
											enchanter[l][2].mana_to_pay = enchanter[l][2].mana_to_pay + unit[j][2][k][2].cost
										else
											unit[j][2][unit[j][2][k][2].name] = nil
											unit[j][2][unit[j][2][k][2].name .. "_level"] = nil
											unit[j][2][k] = nil
											k = k - 1
										end

									end
								end
								wesnoth.put_unit(enchanter)
							else
								unit[j][2][unit[j][2][k][2].name] = nil
								unit[j][2][unit[j][2][k][2].name .. "_level"] = nil
								unit[j][2][k] = nil
								k = k - 1
							end
						end
					end
				end
				for k = 1,#unit[j][2] do	-- Check if the unit has still some enchantments
					if unit[j][2][k][1] == "enchantment" then
						enchantments = true
					end
				end
			end
		end
		-- If an enchantments collapses, we should remove the enchanted mark to make sure tht the game will not be checking it anymore.
		if enchantments == false then
			for j = 1,#unit do
				if unit[j][1] == "status" then
					unit[j][2].enchanted = nil
				end
			end
		end
		local raw_enchanted = wesnoth.get_units { id=unit.id }	-- For the case if enchanter and enchanted were the same person - now, unit.variables.mana_to_pay can be different, so the unit must be checked again
		local unit2 = raw_enchanted[1].__cfg
		local mana_to_pay
		for j = 1,#unit2 do
			if unit2[j][1] == "variables" then
				mana_to_pay = unit2[j][2].mana_to_pay
			end
		end
		if mana_to_pay then
			for j = 1,#unit do
				if unit[j][1] == "variables" then
					unit[j][2].mana_to_pay = mana_to_pay
				end
			end
			unit = wesnoth.display_mana(unit)
		end
		wesnoth.put_unit(unit)
	end
	-- Check out for cooldown periods of spells that might be ended and the spells might have to be restored, add mana regenerated and subtract the mana_to_pay variable after that
	if side_number <= 5 then
		local raw_unit = wesnoth.get_units { side=side_number, canrecruit=true}
		if #raw_unit > 0 then
			local unit = raw_unit[1].__cfg
			if unit then
				for j = 1,#unit do
					if unit[j][1] == "variables" then
						-- Calculate mana changes
						unit[j][2].mana = unit[j][2].mana + unit[j][2].mana_regen - unit[j][2].mana_to_pay
						unit[j][2].mana_to_pay = 0
						if unit[j][2].max_mana < unit[j][2].mana then
							unit[j][2].mana = unit[j][2].max_mana
						end
						-- Restore attacks whose cooldown periods might have ended
						for k = 1,#unit[j][2] do
							if unit[j][2][k][1] == "cooldowns" then
								local cooldowns = unit[j][2][k][2]
								for m,n in pairs(cooldowns) do
									unit[j][2][k][2][m] = unit[j][2][k][2][m] - 1
									if unit[j][2][k][2][m] == 0 then
										unit[j][2][k][2][m] = nil
									end
								end
							end
						end
					end
				end
				unit = wesnoth.display_mana(unit)
				wesnoth.put_unit(unit)
			end
		end
	end
end

function wesnoth.wml_actions.check_for_too_many_minions(cfg)
	-- This function is called after summoning something, it checks if a summoner has more minions of a kind he may have, so the summons with the lowest hp are killed
	local side_checked = cfg.side_checked or wesnoth.get_variable("side_number")
	-- Find who is the summoner
	local raw_caster = wesnoth.get_units { side=side_checked, canrecruit=true }
	local caster = raw_caster[1].__cfg
	local max_numbers = {}
	for j = 1,#caster do
		if caster[j][1] == "variables" then
			for k = 1,#caster[j][2] do
				if caster[j][2][k][1] == "summons" then
					max_numbers = caster[j][2][k][2]
				end
			end
		end
	end
	-- List the summons that are marked to be removal
	local eliminated = {}
	local found_something = true
	local raw_units = wesnoth.get_units { side=side_checked, {"filter_wml" , {  { "variables" , {  is_summon = true }}  }} }
	while found_something == true do
		-- We will find the one with the lowest amounts of hitpoints
		local minimum_hitpoints = {}
		local minimum_hitpoints_ids = {}
		local numbers = {}
		found_something = false
		raw_units = wesnoth.get_units { side=side_checked, {"filter_wml" , {  { "variables" , {  is_summon = true }}  }} }
		-- Count them
		for i = 1,#raw_units do
			local unit = raw_units[i].__cfg
			local summon_data = {}
			local skip = false
			for j,k in pairs(eliminated) do
				-- We skip those who are considered to be deleted already
				if k == unit.id then
					skip=true
				end
			end
			if skip == false then
				for j = 1,#unit do
					-- We check how many of them the caster may have, if he can have more than one
					if unit[j][1] == "variables" then
						for k = 1,#unit[j][2] do
							if unit[j][2][k][1] == "summon_data" then
								summon_data = unit[j][2][k][2]
								if numbers[summon_data.skill] then
									numbers[summon_data.skill] = numbers[summon_data.skill] + 1
								else
									numbers[summon_data.skill] = 1
								end
							end
						end
					end
				end
				-- We check if this unit has the lowest number of hitpoints
				if minimum_hitpoints[summon_data.skill] then
					if minimum_hitpoints[summon_data.skill] > unit.hitpoints then
						minimum_hitpoints[summon_data.skill] = unit.hitpoints
						minimum_hitpoints_ids[summon_data.skill] = unit.id
					end
				else
					minimum_hitpoints[summon_data.skill] = unit.hitpoints
					minimum_hitpoints_ids[summon_data.skill] = unit.id
				end
			end
		end
		-- Remove the excessive ones
		for i,j in pairs(minimum_hitpoints_ids) do
			if max_numbers["summon_" .. i .. "_max"] then
				if max_numbers["summon_" .. i .. "_max"] < numbers[i] then
					eliminated[#eliminated+1] = j
					found_something = true
					wesnoth.wml_actions.kill{ id = j, fire_event = false}
				end
			else
				eliminated[#eliminated+1] = j
				found_something = true
				wesnoth.wml_actions.kill{ id = j, fire_event = false}
			end
		end
		-- Because it is a while cycle, the remaining units will be eliminated in another repeating if the summoner has much more minions than he should for some reasons
	end
end

function wesnoth.wml_actions.status_timeout(cfg)
	-- Another function fired by WML in turn refresh. It makes some statuses disappear with a certain half-life decay, and remove the objects that hold their effects if they exist
	local side_number = wesnoth.get_variable("side_number")
	local status_name = cfg.name or helper.wml_error("[status_timeout] missing required name= attribute")
	local chance = cfg.chance or helper.wml_error("[status_timeout] missing required chance= attribute")
	-- Find the units with the status
	local raw_units = wesnoth.get_units { side=side_number, {"filter_wml" , {  { "status" , {  [status_name]=true }}  }} }
	for i = 1,#raw_units do
		local unit = raw_units[i].__cfg
		local statuses = false	-- We have to be careful, the unit can have the status more than once, as some statuses can stack with itself
		for j = 1,#unit do
			if unit[j][1] == "modifications" then
				for k = 1,#unit[j][2] do
					if unit[j][2][k][1] == "object" then
						-- Find the [object] tied to the status
						if unit[j][2][k][2].description == status_name then
							-- Decide if it's time for it to end
							local roll = helper.rand("1..100")
							if roll <= chance then
								unit[j][2][k] = nil
								k = k - 1
							end
						end
					end
				end
				-- Check if there are still other effects producing the status
				for k = 1,#unit[j][2] do
					if unit[j][2][k][1] == "object" then
						if unit[j][2][k][2].description == status_name then
							statuses = true
						end
					end
				end
			end
		end
		-- Remove the status if the all objects are decayed
		if statuses == false then
			for j = 1,#unit do
				if unit[j][1] == "status" then
					unit[j][2][status_name] = nil
				end
			end
		end
		wesnoth.put_unit(unit)
		wesnoth.wml_actions.reload_unit_modifications{ x=unit.x, y=unit.y }
	end
end

function wesnoth.wml_actions.status_timeout_simple(cfg)
	-- This is a simplified version of the funciton above, for statuses that don't have events, and the statuses are simply marks that an even will be fired upon these units - like specific poisons
	local side_number = wesnoth.get_variable("side_number")
	local status_name = cfg.name or helper.wml_error("[status_timeout] missing required name= attribute")
	local chance = cfg.chance or helper.wml_error("[status_timeout] missing required chance= attribute")
	local raw_units = wesnoth.get_units { side=side_number, {"filter_wml" , {  { "status" , {  [status_name]=true }}  }} }
	for i = 1,#raw_units do
		local unit = raw_units[i].__cfg
		for j = 1,#unit do
			-- Here we just check if the status will be removed or not
			if unit[j][1] == "status" then
				local roll = helper.rand("1..100")
				if roll <= chance then
					unit[j][2][status_name] = nil
				end
			end
		end
		wesnoth.put_unit(unit)
	end
end

-- Useful function taken from elsewhere, can be done also by copying it to WML and getting back, though this is probably faster
function wesnoth.deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
	copy = {}
	for orig_key, orig_value in next, orig, nil do
	    copy[wesnoth.deepcopy(orig_key)] = wesnoth.deepcopy(orig_value)
	end
	setmetatable(copy, wesnoth.deepcopy(getmetatable(orig)))
    else -- number, string, boolean, and so on
	copy = orig
    end
    return copy
end


-- Taken from elsewhere for debugging - it is crucial to know what is contained in a variable if it is generated by the engine and not by this lua code, and also easier to use it to check what is in the variable
-- than read through all the functions that created it (when editing an older code)
debug_utils = {}

-- This is an extensive debug message function
-- It outputs information about type, value, length, and metatable of a variable of any lua kind.
-- It also distinguishes between general tables and tables which are also wml tables. (wml tables are a subamount of tables)
-- If a table isn't a wml table it displays info about the reason.
-- If the variable is a table or if it can be dumped to a table an extensive syntactically correct output is displayed.
-- arguments:
-- lua_var: the variable to investigate
-- clear (boolean, optional): whether the chat window should be cleared before displaying the message
-- name (string, optional): a name to be assigned to the variable, used in the message; useful if there are several
-- variables to be outputted to distinguish their messages
-- onscreen (boolean, optional): whether the message shall be displayed in a wml [message] dialog too
-- That [message] dialog can get very slow for large tables such as unit arrays.
function wesnoth.dbms(lua_var, clear, name, onscreen, wrap, only_return)
	if type(clear) ~= "boolean" then clear = true end
	if type(name) ~= "string" then name = "lua_var" end
	if type(onscreen) ~= "boolean" then onscreen = true end

	local function dump_userdata(data)
		local metatable = getmetatable(data)
		if metatable == "side" then return data.__cfg, true end
		if metatable == "unit" then return data.__cfg, true end
		local data_to_string = tostring(data)
		if metatable == "translatable string" then return data_to_string, false end
		if metatable == "wml object" then return data.__literal, false end
		return data_to_string, true
	end

	local is_wml_table = true
	local result
	local wml_table_error
	local base_indent = "    "
	local function table_to_string(arg_table, indent, introduces_subtag, indices)
		local is_filled = false

		local function check_subtag()
			local one, two = arg_table[1], arg_table[2]
			if type(two) == "userdata" then
				local invalidate_wml_table
				two, invalidate_wml_table = dump_userdata(two)
				if invalidate_wml_table then return false end
			end
			if type(one) ~= "string" or type(two) ~= "table" then return false end
			return true
		end
		if is_wml_table and introduces_subtag then
			if not check_subtag() then
				wml_table_error = string.format("table introducing subtag at %s is not of the form {\"tag_name\", {}}", indices)
				is_wml_table = false
			end
		end

		local index = 1
		for current_key, current_value in pairs(arg_table) do
			is_filled = true

			local current_key_type = type(current_key)
			local current_key_to_string = tostring(current_key)
			local current_type = type(current_value)
			local function no_wml_table(expected, index, type)
				if not index then index = current_key_to_string end
				if not type then type = current_type end
				wml_table_error = string.format("value at %s[%s]: %s expected, got %s", indices, index, expected, type)
				is_wml_table = false
			end

			if current_type == "userdata" then
				local  invalidate_wml_table
				current_value, invalidate_wml_table = dump_userdata(current_value)
				current_type = type(current_value)
				if is_wml_table and invalidate_wml_table then
					wml_table_error = string.format("userdata at %s[%s]", indices, current_key_to_string)
					is_wml_table = false
				end
			end

			if is_wml_table and not introduces_subtag then
				if current_key_type == "string" and (current_type == "table" or current_type == "function" or current_type == "thread") then
					no_wml_table("nil, boolean, number or string")
				elseif current_key_type == "number" then
					if current_type ~= "table" then
						no_wml_table("table")
					elseif current_key ~= index then
						no_wml_table("value", tostring(index), "nil or fields traversed out-of-order")
					end
					index = index + 1
				end
			end

			local length = 9
			local left_bracket, right_bracket = "[", "]"
			if current_key_type == "string" then
				left_bracket, right_bracket = "", ""; length = length - 2
			end
			if current_type == "table" then
				result = string.format("%s%s%s%s%s = {\n", result, indent, left_bracket, current_key_to_string, right_bracket)
				table_to_string(current_value, string.format("%s%s%s", base_indent, indent, string.rep(" ", string.len(current_key_to_string) + length)),
					not introduces_subtag, string.format("%s[%s]", indices, current_key_to_string))
				result = string.format("%s%s%s},\n", result, indent, string.rep(" ", string.len(current_key_to_string) + length))
			else
				local quote = ""; if current_type == "string" then quote = "\"" end
				result = string.format("%s%s%s%s%s = %s%s%s,\n", result, indent, left_bracket, current_key_to_string, right_bracket, quote, tostring(current_value), quote)
			end
		end
		if is_filled then result = string.sub(result, 1, string.len(result) - 2) .. "\n" end
	end

	local engine_is_wml_table
	if wesnoth then
		engine_is_wml_table = pcall(wesnoth.set_variable, "LUA_debug_msg", lua_var); wesnoth.set_variable("LUA_debug_msg")
	end

	local var_type, var_value = type(lua_var), tostring(lua_var)
	is_wml_table = var_type == "table"
	local invalidate_wml_table
	local metatable = getmetatable(lua_var)
	if var_type == "userdata" then
		lua_var, invalidate_wml_table= dump_userdata(lua_var)
		is_wml_table = not invalidate_wml_table
	end
	local new_var_type = type(lua_var)

	local format_string = "%s is of type %s, value %s"
	local format_string_length = format_string .. ", length %u"
	local format_string_length_newline = format_string_length .. ":\n%s"

	if new_var_type == "table" then
		local var_length = #lua_var
		result = "{\n"
		table_to_string(lua_var, base_indent, false, "")
		result = result .. "}"

		if is_wml_table then
			result = string.format(format_string_length_newline, name, "WML table", var_value, var_length, result)
		elseif wml_table_error then
			result = string.format(format_string_length .. ", but no WML table: %s:\n%s", name, "table", var_value, var_length, wml_table_error, result)
		else
			result = string.format(format_string_length_newline, name, var_type, var_value, var_length, result)
		end
	elseif new_var_type == "string" then
		result = string.format(format_string_length, name, var_type, var_value, #lua_var)
	else
		result = string.format(format_string, name, var_type, var_value)
	end

	if metatable then result = string.format("%s\nwith a metatable:\n", result) end

	if wesnoth and is_wml_table ~= engine_is_wml_table  and (var_type == "table" or var_type == "userdata" or var_type == "function" or var_type == "thread") then
		result = string.format("warning: WML table inconsistently predicted, script says %s , engine %s \n%s", tostring(is_wml_table), tostring(engine_is_wml_table), result)
	end

	if clear and wesnoth then wesnoth.clear_messages() end
	if not only_return then
--		if wesnoth then wesnoth.message("dbms", result) end
 print(result)
	end
	local continue = true
	if onscreen and wesnoth and not only_return then
		if wrap then wesnoth.wml_actions.message({ speaker = "narrator", image = "wesnoth-icon.png", message = result })
		else
			local wlp_utils = wesnoth.require "~add-ons/Wesnoth_Lua_Pack/wlp_utils.lua"
			local result = wlp_utils.message({ caption = "dbms", message = result })
			if result == -2 then continue = false end
		end
	end
	if metatable and continue then
		result = result .. debug_utils.dbms(metatable, false, string.format("The metatable %s", tostring(metatable)), onscreen, wrap, only_return)
	end
	return result
end

